﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  [Item("PointCertaintyWeightCalculator", "")]
  public class PointCertaintyWeightCalculator : DiscriminantClassificationWeightCalculator {

    public PointCertaintyWeightCalculator()
      : base() {
    }
    [StorableConstructor]
    protected PointCertaintyWeightCalculator(bool deserializing) : base(deserializing) { }
    protected PointCertaintyWeightCalculator(PointCertaintyWeightCalculator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new PointCertaintyWeightCalculator(this, cloner);
    }

    protected override IEnumerable<double> DiscriminantCalculateWeights(IEnumerable<IDiscriminantFunctionClassificationSolution> discriminantSolutions) {
      List<double> weights = new List<double>();
      IClassificationProblemData problemData = discriminantSolutions.ElementAt(0).ProblemData;
      // class Values are the same in all problem data sets
      double avg = problemData.ClassValues.Average();

      IEnumerable<double> targetValues;
      IEnumerator<double> trainingValues;

      foreach (var solution in discriminantSolutions) {
        problemData = solution.ProblemData;
        targetValues = GetValues(problemData.Dataset.GetDoubleValues(problemData.TargetVariable).ToList(), problemData.TrainingIndizes);
        trainingValues = targetValues.GetEnumerator();

        IEnumerator<double> estimatedTrainingVal = solution.EstimatedTrainingValues.GetEnumerator();
        IEnumerator<double> estimatedTrainingClassVal = solution.EstimatedTrainingClassValues.GetEnumerator();

        double curWeight = 0.0;
        while (estimatedTrainingVal.MoveNext() && estimatedTrainingClassVal.MoveNext() && trainingValues.MoveNext()) {
          if (estimatedTrainingClassVal.Current.Equals(trainingValues.Current)) {
            curWeight += 0.5;
            double distanceToPoint = Math.Abs(estimatedTrainingVal.Current - avg);
            double distanceToClass = Math.Abs(trainingValues.Current - avg);
            if (trainingValues.Current > avg && estimatedTrainingVal.Current > avg
             || trainingValues.Current < avg && estimatedTrainingVal.Current < avg)
              curWeight += distanceToPoint < distanceToClass ? (0.5 / distanceToClass) * distanceToPoint : 0.5;
          }
        }
        // normalize the weight (otherwise a model with a bigger training partition would probably be better)
        weights.Add(curWeight / targetValues.Count());
      }
      return weights;
    }
  }
}
